#include	<conio.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<sys/scsi.h>
#include	"scsisub.h"
#include	"version.h"

#define	RETRY_COUNT	3
#define	SENSE_BUFFER_SIZE	48
#ifndef	DEFAULT_LUN
  #define	DEFAULT_LUN	0
#endif
#ifndef	MULTI_BLOCKS
  #define	MULTI_BLOCKS	32
#endif

/* XCb`w */
int opt_c = 0;
int opt_a = 0;
int opt_r = 0;
int opt_w = 0;
int opt_f = 0;
int verbose = 0;

/* g̖O */
char *progname;


void version(void)
{
	fprintf(stderr, "%s %s (c) JACK\n", progname, VERSION);
}

void usage(void)
{
	fprintf(stderr, "usage: %s [-carwfvhV] <SCSI-ID> [<start block>]\n", progname);
}

void help(void)
{
	static char *help_message[] = {
		" -c        verify check",
		" -a        auto reassign block",
		" -r        reassign block",
		" -w        sector write check",
		" -f        format unit",
		" -v        verbose mode",
		" -h        give this help",
		" -V        display version number\n",
		NULL
	};
	char **p = help_message;

	version();
	usage();
	while(*p)	fprintf(stderr, "%s\n", *p++);
}

void *xmalloc(size_t size)
{
	void *p;
	if ((p = malloc(size)))	return p;
	perror(NULL);
	exit(EXIT_FAILURE);
}

int ynget(void)
{
	int c;

	while (1) {
		if (kbhit()) {
			c = getch();
			if (c == 0x1b)	break;
			if (c == 'y' || c == 'Y')	return 1;
			if (c == 'n' || c == 'N')	return 0;
		}
	}
	return 0;
}

char *get_sense_error_message(unsigned char code, unsigned char key)
{
	static char *sense_error_message[] = {
						"m̃G[܂.",
						"G[͂܂.",
						"CfbNX/ZN^Moł܂.",
						"V[NɏIł܂.",
						"CgɎs܂.",
						"ȕłĂ܂.",
						"gbNO܂.",
						"CgɃItEgbN܂.",
						"hc bqbG[܂.",
						"hc̓ǂݍ݂ł܂.",
						"gCɂǂݍ݃G[̉񕜂ł܂.",
						"hc̃AhX}[Noł܂.",
						"f[^̃AhX}[Noł܂.",
						"ubNȂ.",
						"V[Nňُ킪܂.",
						"gCɂf[^ǂݏo܂.",
						"dbbɂf[^ǂݏo܂.",
						"׃Xg̓ǂݏoł܂.",
						"p[^]ɃI[o[E܂.",
						"ُȃR}hR[h󂯎܂.",
						"ُȃubNw肳܂.",
						"bcaɈُȒlw肳܂.",
						"ُȂktmw肳܂.",
						"p[^XgɈُȒlw肳܂.",
						"CgEveNgĂ܂.",
						"uԂɂȂĂ.",
						"[hEZNgEp[^ύXꂽ.",
						"}ׂ̌ɂtH[}bgɎs܂.",
						"׃ubNɊ蓖ĂXyÄ悪܂.",
						"ffAobt@q`lɈُ킪܂.",
						"ffAdbbHɈُ킪܂.",
						"d̎Ȑffňُ킪܂.",
						"bZ[WEWFNg󂯎܂.",
						"ZNV/ZNVɎs܂.",
						"peBEG[܂.",
						"CjVG[^EfBeNgEG[󂯎܂.",
						"ُȃbZ[W󂯎܂.",
	};

	switch (code) {
		case 0x00:
			return sense_error_message[1];
		case 0x01:
			return sense_error_message[2];
		case 0x02:
			return sense_error_message[3];
		case 0x03:
			return sense_error_message[4];
		case 0x04:
			return sense_error_message[5];
		case 0x06:
			return sense_error_message[6];
		case 0x09:
			return sense_error_message[7];
		case 0x10:
			switch (key) {
				case 0x03:
					return sense_error_message[8];
				case 0x04:
					return sense_error_message[9];
			}
			break;
		case 0x11:
			return sense_error_message[10];
		case 0x12:
			return sense_error_message[11];
		case 0x13:
			return sense_error_message[12];
		case 0x14:
			return sense_error_message[13];
		case 0x15:
			return sense_error_message[14];
		case 0x17:
			return sense_error_message[15];
		case 0x18:
			return sense_error_message[16];
		case 0x19:
			return sense_error_message[17];
		case 0x1a:
			return sense_error_message[18];
		case 0x20:
			return sense_error_message[19];
		case 0x21:
			return sense_error_message[20];
		case 0x24:
			return sense_error_message[21];
		case 0x25:
			return sense_error_message[22];
		case 0x26:
			return sense_error_message[23];
		case 0x27:
			return sense_error_message[24];
		case 0x29:
			return sense_error_message[25];
		case 0x2a:
			return sense_error_message[26];
		case 0x31:
			return sense_error_message[27];
		case 0x32:
			return sense_error_message[28];
		case 0x40:
			return sense_error_message[29];
		case 0x41:
			return sense_error_message[30];
		case 0x42:
			return sense_error_message[31];
		case 0x43:
			return sense_error_message[32];
		case 0x45:
			return sense_error_message[33];
		case 0x47:
			return sense_error_message[34];
		case 0x48:
			return sense_error_message[35];
		case 0x49:
			return sense_error_message[36];
	}
	return sense_error_message[0];
}

void print_sense_data(unsigned char *sense_buffer, int sense_size)
{
	int i,j, max;
	unsigned char code, key;

	if ((sense_buffer[0] & 0x70) == 0x70) {
		/* gZXf[^ */
		code = sense_buffer[12];
		key = sense_buffer[2] & 0x0f;
		printf("%s",get_sense_error_message(code, key));
		if (verbose) {
			printf("(CODE=0x%2.2x, KEY=0x%2.2x)", code, key);
		}
		printf("\n");
	} else {
		code = sense_buffer[0] & 0x7f;
		key = 0;
		printf("ZXER[h = 0x%2.2x\n", code);
	}
	if (verbose) {
		printf("ZXEf[^\n");
		for (i = 0; i < sense_size; i += 16) {
			max = i+16 > sense_size ? sense_size - i : 16;
			for (j = 0; j < max; j++) {
				printf("%2.2x ", sense_buffer[i+j]);
			}
			printf("\n");
		}
	}
}

void reassign_block(int id, int lun, int block)
{
	unsigned char sense_buffer[SENSE_BUFFER_SIZE];
	int ret;

	if ((ret = reassign(id, lun, block)) != 0) {
		if (ret == -1) {
			printf("ATCEubNXER}h̎sł܂ł.\n");
		} else if (ret == -2) {
			printf("SCSIoXɃf[^o܂ł.\n");
		} else if ((ret & 0x00ff) == 0x02) {
			if (sense(id, lun, sense_buffer, sizeof(sense_buffer)) >= 0) {
				print_sense_data(sense_buffer, sizeof(sense_buffer));
			}
		} else {
			printf("̑̃G[ %8.8x\n", ret);
		}
	} else {
		printf("܂.\n");
	}
}

void check_sectors(int id, int lun, struct _readcap *cap, int block)
{
	unsigned char sense_buffer[SENSE_BUFFER_SIZE];
	unsigned char *read_buffer;
	unsigned char *write_buffer;
	int i = 0, j = 0, max = 0;
	int cancel = 0;
	int c;
	int rret;
	int wret;

	printf("\n");
	if (opt_c) {
		if (opt_w) {
			printf("read & write(verify) check ...\n");
		} else {
			printf("verify check ...\n");
		}
	} else {
		if (opt_w) {
			printf("read & write check ...\n");
		} else {
			printf("read check ...\n");
		}
	}
	read_buffer = write_buffer = xmalloc(cap->size * MULTI_BLOCKS);
	while (kbhit()) {
		getch();
	}
	for (i = block; i <= cap->block; i += MULTI_BLOCKS) {
		if (kbhit() && (c = getch())) {
			if (c == 0x1b) {
				if (verbose) {
					printf("\n");
				}
				printf("LZ܂.\n");
				cancel = 1;
				break;
			}
		}
		max = (i + MULTI_BLOCKS-1) >= cap->block+1 ? cap->block+1 - i : MULTI_BLOCKS;
		if (max == 0)	break;
		if (verbose) {
			printf("ubN %d", i);
			fflush(stdout);
		}
		if ((rret = (opt_c && (!opt_w)) ?
		verify(id, lun, i, max) :
		readextended(id, lun, i, max, cap->size, read_buffer)) != 0 ||
		(opt_w && (wret = opt_c ?
		writeverify(id, lun, i, max, cap->size, write_buffer) :
		writeextended(id, lun, i, max, cap->size, write_buffer)) != 0)) {
			if (verbose) {
				printf("\r");
				fflush(stdout);
			}
			for (j = 0; j < max; j++) {
				if ((rret = (opt_c && (!opt_w)) ?
				verify(id, lun, i+j, 1) :
				readextended(id, lun, i+j, 1, cap->size, read_buffer)) != 0) {
					printf("ubN %d: ", i+j);
					fflush(stdout);
					if (rret == -1) {
						if (opt_c && (!opt_w)) {
							printf("xt@CE");
						} else {
							printf("[hE");
						}
						printf("R}h̎sł܂ł.\n");
					} else if (rret == -2) {
						printf("SCSIoX̃f[^܂ł.\n");
					} else if ((rret & 0x00ff) == 0x02) {
						if (sense(id, lun, sense_buffer, sizeof(sense_buffer)) >= 0) {
							print_sense_data(sense_buffer, sizeof(sense_buffer));
						}
					} else {
						printf("̑̃G[ %8.8x\n", rret);
					}
					if (opt_a) {
						printf("׃ubN̍Ĕzus܂. ");
						reassign_block(id, lun, i+j);
					} else if (opt_r) {
						printf("׃ubN̍Ĕzus܂H ");
						fflush(stdout);
						if (ynget()) {
							printf("s܂. ");
							reassign_block(id, lun, i+j);
						} else {
							printf("܂.\n");
						}
					}
				} else if (opt_w && (wret = opt_c ?
				writeverify(id, lun, i+j, 1, cap->size, write_buffer) :
				writeextended(id, lun, i+j, 1, cap->size, write_buffer)) != 0) {
					printf("ubN %d: ", i+j);
					fflush(stdout);
					if (wret == -1) {
						if (opt_c && (!opt_w)) {
							printf("CgEAhExt@CE");
						} else {
							printf("CgE");
						}
						printf("R}h̎sł܂ł.\n");
					} else if (wret == -2) {
						printf("SCSIoXɃf[^o܂ł.\n");
					} else if ((wret & 0x00ff) == 0x02) {
						if (sense(id, lun, sense_buffer, sizeof(sense_buffer)) >= 0) {
							print_sense_data(sense_buffer, sizeof(sense_buffer));
						}
					} else {
						printf("̑̃G[ %8.8x\n", wret);
					}
					if (opt_a) {
						printf("׃ubN̍Ĕzus܂. ");
						reassign_block(id, lun, i+j);
					} else if (opt_r) {
						printf("׃ubN̍Ĕzus܂H ");
						fflush(stdout);
						if (ynget()) {
							printf("s܂. ");
							reassign_block(id, lun, i+j);
						} else {
							printf("܂.\n");
						}
					}
				}
			}
		}
		if (verbose) {
			printf("\r");
			fflush(stdout);
		}
	}
	if (verbose) {
		if (!cancel) {
			printf("ubN %d", (i - MULTI_BLOCKS + (max > 0 ? max-1 : 0)));
		}
		printf("\n");
	}
	fflush(stdout);
	free(read_buffer);
	free(write_buffer);
}

void format_unit(int id, int lun)
{
	unsigned char sense_buffer[SENSE_BUFFER_SIZE];
	int ret;

	printf("format unit ...\n");
	if ((ret = format(id, lun)) != 0) {
		if (ret == -1) {
			printf("tH[}bgER}h̎sł܂ł.\n");
		} else if ((ret & 0x00ff) == 0x02) {
			if (sense(id, lun, sense_buffer, sizeof(sense_buffer)) >= 0) {
				print_sense_data(sense_buffer, sizeof(sense_buffer));
			}
		} else {
			printf("̑̃G[ %8.8x\n", ret);
		}
	} else {
		printf("܂.\n");
	}
}

int main(int argc, char **argv)
{
	struct _readcap *cap;
	int ch;
	int id;
	int block;
	int ret;
	int maxsize;

	progname = argv[0];

	while ((ch = getopt(argc, argv, "carwfvhV")) != EOF) {
		switch (ch) {
			case 'c':
				opt_c = 1; break;
			case 'a':
				opt_a = 1; opt_r = 0; break;
			case 'r':
				opt_r = 1; opt_a = 0; break;
			case 'w':
				opt_w = 1; break;
			case 'f':
				opt_f = 1; break;
			case 'v':
				verbose = 1; break;
			case 'h':
				help(); exit(EXIT_FAILURE); break;
			case 'V':
				version(); exit(EXIT_SUCCESS); break;
			default:
				usage(); exit(EXIT_FAILURE); break;
		}
	}

	if (optind >= argc) {
		usage();
		exit(EXIT_FAILURE);
	} else {
		id = atoi(argv[optind]);
		if (id < 0 || id > 7) {
			printf("SCSI-IDw肵Ă.\n");
			exit(EXIT_FAILURE);
		}
	}
	if (optind+1 < argc) {
		block = atoi(argv[optind+1]);
	} else {
		block = 0;
	}

	printf("^[Qbg SCSI-ID  :%d\n", id);
	if (!opt_f) {
		printf("`FbNJnubN:%d\n", block);
	}

	/* ԃ`FbN */
	if (bus_check() != 0) {
		printf("SCSIoXgpĂ\܂.\n");
		return -1;
	}
	while ((ret = testunit(id, DEFAULT_LUN)) != 0) {
		if (ret == -1) {
			printf("ȕłĂȂ悤ł.\n");
			exit(EXIT_FAILURE);
		}
	}

	if (opt_f) {
		format_unit(id, DEFAULT_LUN);
		return 0;
	}

	/* }̂̃TCY`FbN */
	cap = xmalloc(sizeof(struct _readcap));
	readcapacity(id, DEFAULT_LUN, (unsigned char*)cap);
	printf("őubN  : %d\n", cap->block);
	printf("ubNTCY: %d\n", cap->size);
	maxsize = 1024/cap->size;
	maxsize *= (cap->block+1);
	printf("}̗e      : %d(KB)\n", maxsize);

	if (block > cap->block) {
		printf("`FbNJnubNőubN𒴂Ă܂.\n");
		exit(EXIT_FAILURE);
	}
	check_sectors(id, DEFAULT_LUN, cap, block);
	free(cap);
	return 0;
}
