#include	<sys/scsi.h>
#include	"scsisub.h"

#define	RETRY_COUNT	3

int bus_check(void)
{
	int i;

	/* SCSIoX̃`FbN */
	for (i = 0; i < RETRY_COUNT; i++) {
		if (_scsi_phase() == 0) {
			return 0;
		}
	}
	return -1;
}

int command_out(int id, unsigned char *command_buffer, int size)
{
	int i;

	for (i = 0; i < RETRY_COUNT; i++) {
		/* ZNVtF[Y */
		if (_scsi_select(id) == 0) {
			/* R}htF[Y */
			if (_scsi_cmdout(size, command_buffer) == 0) {
				return 0;
			} else {
				return -1;
			}
		}
	}
	return -1;
}

int message_get(void)
{
	unsigned char status;
	unsigned char message;
	int ret;

	/* Xe[^XtF[Y */
	if ((ret = _scsi_stsin(&status)) != 0) {
		return -1;
	}
	/* bZ[WtF[Y */
	if (_scsi_msgin(&message) != 0) {
		return -1;
	}
	return ((int)(message & 0xff) << 8) + ((int)(status & 0x8f));
}

int testunit(int id, int lun)
{
	unsigned char command_buffer[6];
	int message;

	command_buffer[0] = 0x00;				/* eXgEjbg */
	command_buffer[1] = ((lun & 0x7) << 5);	/* uio */
	command_buffer[2] = 0;
	command_buffer[3] = 0;
	command_buffer[4] = 0;
	command_buffer[5] = 0;

	if (command_out(id, command_buffer, 6) != 0) {
		return -1;
	}
	message = message_get();
	return message;
}

int rezerounit(int id, int lun)
{
	unsigned char command_buffer[6];
	int message;

	command_buffer[0] = 0x01;				/* [Ejbg */
	command_buffer[1] = ((lun & 0x7) << 5);	/* uio */
	command_buffer[2] = 0;
	command_buffer[3] = 0;
	command_buffer[4] = 0;
	command_buffer[5] = 0;

	if (command_out(id, command_buffer, 6) != 0) {
		return -1;
	}
	message = message_get();
	return message;
}

int sense(int id, int lun, unsigned char *sense_buffer, int sense_size)
{
	unsigned char command_buffer[6];
	int message;

	command_buffer[0] = 0x03;				/* NGXgEZX */
	command_buffer[1] = ((lun & 0x7) << 5);	/* uio */
	command_buffer[2] = 0;
	command_buffer[3] = 0;
	command_buffer[4] = sense_size & 0xff;	/* AP[VEOX */
	command_buffer[5] = 0;

	if (command_out(id, command_buffer, 6) != 0) {
		return -1;
	}
	if (_scsi_datain((sense_size & 0xff), sense_buffer) != 0) {
		message_get();
		return -2;
	}
	message = message_get();
	return message;
}

int format(int id, int lun)
{
	unsigned char command_buffer[6];
	int message;

	command_buffer[0] = 0x04;				/* tH[}bgEjbg */
	command_buffer[1] = ((lun & 0x7) << 5);
	command_buffer[2] = 0;
	command_buffer[3] = 0;
	command_buffer[4] = 0;
	command_buffer[5] = 0;

	if (command_out(id, command_buffer, 6) != 0) {
		return -1;
	}
	message = message_get();
	return message;
}

int reassign(int id, int lun, int block)
{
	unsigned char command_buffer[6];
	unsigned char reassign_list[8];
	int message;

	command_buffer[0] = 0x07;				/* ATCEubN */
	command_buffer[1] = ((lun & 0x7) << 5);	/* uio */
	command_buffer[2] = 0;
	command_buffer[3] = 0;
	command_buffer[4] = 0;
	command_buffer[5] = 0;

	reassign_list[0] = 0;
	reassign_list[1] = 0;
	reassign_list[2] = 0;
	reassign_list[3] = 4;
	reassign_list[4] = (block >> 24) & 0x000000ff;
	reassign_list[5] = (block >> 16) & 0x000000ff;
	reassign_list[6] = (block >>  8) & 0x000000ff;
	reassign_list[7] = (block >>  0) & 0x000000ff;

	if (command_out(id, command_buffer, 6) != 0) {
		return -1;
	}
	if (_scsi_dataout(8, reassign_list) != 0) {
		message_get();
		return -2;
	}
	message = message_get();
	return message;
}

int readcapacity(int id, int lun, unsigned char *capacity_buffer)
{
	unsigned char command_buffer[10];
	int message;

	command_buffer[0] = 0x25;				/* [hELpVeB */
	command_buffer[1] = ((lun & 0x7) << 5);	/* uio */
	command_buffer[2] = 0;
	command_buffer[3] = 0;
	command_buffer[4] = 0;
	command_buffer[5] = 0;
	command_buffer[6] = 0;
	command_buffer[7] = 0;
	command_buffer[8] = 0;
	command_buffer[9] = 0;

	if (command_out(id, command_buffer, 10) != 0) {
		return -1;
	}
	if (_scsi_datain(8, capacity_buffer) != 0) {
		message_get();
		return -2;
	}
	message = message_get();
	return message;
}

int readextended(int id, int lun, int block, int count, int blocksize, unsigned char *read_buffer)
{
	unsigned char command_buffer[10];
	unsigned char *p;
	int message;
	int i;

	command_buffer[0] = 0x28;				/* [hEGNXefBbh */
	command_buffer[1] = ((lun & 0x7) << 5);	/* uio */
	command_buffer[2] = (block >> 24) & 0x000000ff;
	command_buffer[3] = (block >> 16) & 0x000000ff;
	command_buffer[4] = (block >>  8) & 0x000000ff;
	command_buffer[5] = (block >>  0) & 0x000000ff;
	command_buffer[6] = 0;
	command_buffer[7] = (count >>  8) & 0x000000ff;
	command_buffer[8] = (count >>  0) & 0x000000ff;
	command_buffer[9] = 0;

	if (command_out(id, command_buffer, 10) != 0) {
		return -1;
	}
	p = read_buffer;
	for (i = 0; i < (count & 0xffff); i++) {
		if (_scsi_datain(blocksize, p) != 0) {
			message_get();
			return -2;
		}
		p += blocksize;
	}
	message = message_get();
	return message;
}

int writeextended(int id, int lun, int block, int count, int blocksize, unsigned char *write_buffer)
{
	unsigned char command_buffer[10];
	unsigned char *p;
	int message;
	int i;

	command_buffer[0] = 0x2a;				/* CgEGNXefBbh */
	command_buffer[1] = ((lun & 0x7) << 5);	/* uio */
	command_buffer[2] = (block >> 24) & 0x000000ff;
	command_buffer[3] = (block >> 16) & 0x000000ff;
	command_buffer[4] = (block >>  8) & 0x000000ff;
	command_buffer[5] = (block >>  0) & 0x000000ff;
	command_buffer[6] = 0;
	command_buffer[7] = (count >>  8) & 0x000000ff;
	command_buffer[8] = (count >>  0) & 0x000000ff;
	command_buffer[9] = 0;

	if (command_out(id, command_buffer, 10) != 0) {
		return -1;
	}
	p = write_buffer;
	for (i = 0; i < (count & 0xffff); i++) {
		if (_scsi_dataout(blocksize, p) != 0) {
			message_get();
			return -2;
		}
		p += blocksize;
	}
	message = message_get();
	return message;
}

int writeverify(int id, int lun, int block, int count, int blocksize, unsigned char *write_buffer)
{
	unsigned char command_buffer[10];
	unsigned char *p;
	int message;
	int i;

	command_buffer[0] = 0x2e;				/* CgEAhExt@C */
	command_buffer[1] = ((lun & 0x7) << 5);	/* uio */
	command_buffer[2] = (block >> 24) & 0x000000ff;
	command_buffer[3] = (block >> 16) & 0x000000ff;
	command_buffer[4] = (block >>  8) & 0x000000ff;
	command_buffer[5] = (block >>  0) & 0x000000ff;
	command_buffer[6] = 0;
	command_buffer[7] = (count >>  8) & 0x000000ff;
	command_buffer[8] = (count >>  0) & 0x000000ff;
	command_buffer[9] = 0;

	if (command_out(id, command_buffer, 10) != 0) {
		return -1;
	}
	p = write_buffer;
	for (i = 0; i < (count & 0xffff); i++) {
		if (_scsi_dataout(blocksize, p) != 0) {
			message_get();
			return -2;
		}
		p += blocksize;
	}
	message = message_get();
	return message;
}

int verify(int id, int lun, int block, int count)
{
	unsigned char command_buffer[10];
	int message;

	command_buffer[0] = 0x2f;				/* xt@C */
	command_buffer[1] = ((lun & 0x7) << 5);	/* uio */
	command_buffer[2] = (block >> 24) & 0x000000ff;
	command_buffer[3] = (block >> 16) & 0x000000ff;
	command_buffer[4] = (block >>  8) & 0x000000ff;
	command_buffer[5] = (block >>  0) & 0x000000ff;
	command_buffer[6] = 0;
	command_buffer[7] = (count >>  8) & 0x000000ff;
	command_buffer[8] = (count >>  0) & 0x000000ff;
	command_buffer[9] = 0;

	if (command_out(id, command_buffer, 10) != 0) {
		return -1;
	}
	message = message_get();
	return message;
}
